@@ -1,7 +1,7 @@
 /* ykclient.h --- Prototypes for Yubikey OTP validation client library.
  *
  * Written by Simon Josefsson <simon@josefsson.org>.
- * Copyright (c) 2006, 2007, 2008, 2009, 2011 Yubico AB
+ * Copyright (c) 2006-2012 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,6 +36,11 @@
 #include <stdint.h>
 #include <string.h>
 
+# ifdef __cplusplus
+extern "C"
+{
+# endif
+
 typedef enum
 {
   /* Official yubikey client API errors. */
@@ -58,7 +63,8 @@ typedef enum
   YKCLIENT_HEX_DECODE_ERROR,
   YKCLIENT_BAD_SERVER_SIGNATURE,
   YKCLIENT_NOT_IMPLEMENTED,
-  YKCLIENT_CURL_PERFORM_ERROR
+  YKCLIENT_CURL_PERFORM_ERROR,
+  YKCLIENT_BAD_INPUT
 } ykclient_rc;
 
 typedef struct ykclient_st ykclient_t;
@@ -86,6 +92,9 @@ extern int ykclient_set_client_b64 (ykclient_t * ykc,
 extern void ykclient_set_url_template (ykclient_t * ykc,
 				       const char *url_template);
 
+extern int ykclient_set_url_templates (ykclient_t * ykc,
+				       size_t num_templates, const char **url_templates);
+
 /* By default the signature returned by the server is verified (modify
    this choice by calling ykclient_set_verify_signature()). */
 extern void ykclient_set_ca_path (ykclient_t * ykc, const char *ca_path);
@@ -114,4 +123,8 @@ extern int ykclient_verify_otp_v2 (ykclient_t * ykc_in,
 				   size_t urlcount,
 				   const char **urls, const char *api_key);
 
+# ifdef __cplusplus
+}
+# endif
+
 #endif
